
/**************************************************/
/*                                                */
/*  ATBinHex Component                            */
/*  Copyright (C) 2006 Alexey Torgashin           */
/*  http://atorg.net.ru                           */
/*                                                */
/**************************************************/

/*$OPTIMIZATION OFF*/ //Delphi 5 cannot compile this with optimization on
/*$BOOLEVAL OFF*/    //Short boolean evaluation required
/*$RANGECHECKS OFF*/ //For assignment compatability between DWORD and Longint

//ATBinHex options:
 //Enable file change notification code.
                //Should be enabled.
 //Enable printing code.
                //Should be enabled.

//{$define PREVIEW} //Enable ATPrintPreview usage.
                    //Should be disabled.
 //Enable searching code.
                 //Should be enabled.
 //Enable workaround for Windows bug: when horiz scrollbar temporary hides,
                 //this may cause both horiz+vertical scrollbars + window border to disappear
                 //completely. This workaround rises users' question: why horiz scrollbar
                 //doesn't hide when lines become short enough.
                 //Should be enabled.

//{$define REGEX} //Enable DIRegEx usage. Used for URL hilighting.
                  //Should be enabled (with DIRegEx).

//{$define TEST} //Show debug form.
                 //Must be disabled in release.
   //ATBinHex options

#pragma resource "ATBinHexResources.res" //ATBinHex resources
#ifndef ATBinHexH
#define ATBinHexH


#include "d2c_system.h"

#include <windows.hpp>
#include <messages.hpp>
#include <sysutils.hpp>
#include <classes.hpp>
#include <controls.hpp>
#include <graphics.hpp>
#include <stdctrls.hpp>
#include <extctrls.hpp>
#include "ATFileNotification.h"
#include "ATFileNotificationSimple.h"
#include "ATStreamSearch.h"
#include <dialogs.hpp>
#include "ATxCodepages.h"
#include <menus.hpp>



struct ATBinHex__0;
class TATBinHex;
struct TATBinHexOutputOptions;




enum TATBinHexMode {vbmodeText,
                    vbmodeBinary,
                    vbmodeHex,
                    vbmodeUnicode,
                    vbmodeUHex };
enum TATUnicodeFormat {vbUnicodeFmtUnknown,
                       vbUnicodeFmtLE,
                       vbUnicodeFmtBE };
enum TATDirection {vdirUp,
                   vdirDown };
enum TATMouseRelativePosition {vmPosInner,
                               vmPosUpper,
                               vmPosLower,
                               vmPosLefter,
                               vmPosRighter };
enum TATFileSource {vfSrcNone,
                    vfSrcFile,
                    vfSrcStream,
                    vfSrcUrl };
enum TATLineType {vbLineAll,
                  vbLineWord,
                  vbLineURL };
enum TATPopupCommand {vpCmdCopy,
                      vpCmdCopyHex,
                      vpCmdCopyLink,
                      vpCmdSelectLine,
                      vpCmdSelectAll,
                      vpCmdEncMenu };
typedef System::Set < TATPopupCommand, vpCmdCopy, vpCmdEncMenu > TATPopupCommands;

struct TATBinHexOutputOptions {      //"Show non-printable" mode is on
                //Current line has CR (not wrapped)
             //Current font is FontOEM, not Font
  bool ShowNonPrintable, ShowCR, IsFontOem, IsFontFixed;  //Current font has fixed width
  int TabSize;      //"Tab size" value
};


typedef void __fastcall ( __closure * TATBinHexDrawLine )( TObject*, TCanvas*, const WideString, const WideString, const TRect&, const TPoint&, bool& );
typedef void __fastcall ( __closure * TATBinHexDrawLine2 )( TObject*, TCanvas*, const WideString, const TPoint&, const TATBinHexOutputOptions& );
typedef void __fastcall ( __closure * TATBinHexClickURL )( TObject*, const AnsiString );
const int cMaxLength = 2 * 1024; //Limits for "Maximal line length" value

const int cMinLength = 2;        //

const int cMaxLengthSel = 8 * 1024; //Max length for "select line" command

const int cMaxURLs = 500;        //Max URLs count in buffer


typedef int TATStringExtent [ 2049/*# range 0..cMaxLength*/ ];
struct ATBinHex__0 {
  AnsiString FString;
  __int64 FPos;
};
typedef ATBinHex__0 TATUrlArray [ 500/*# range 1..cMaxURLs*/ ];
bool __fastcall StringExtent( TCanvas* ACanvas, const WideString AStr, TATStringExtent& AExtent, const TATBinHexOutputOptions& AOptions );
const System::Set<TATPopupCommand, vpCmdCopy, vpCmdEncMenu> cATBinHexCommandSet = ( System::Set < TATPopupCommand, vpCmdCopy, vpCmdEncMenu >() \
                     << TATPopupCommand ( 0 ) << TATPopupCommand ( 1 ) << TATPopupCommand ( 2 ) << TATPopupCommand ( 3 ) << TATPopupCommand ( 4 ) << TATPopupCommand ( 5 ) );
const int cATBinHexBkColor = 0xECECEC;  //Default Hex mode back color


typedef int* PATBinHexIntArray;
typedef int TIntegerArray [ 1000000/*# range 1..1000*1000*/ ];


class TATBinHex: public TPanel {
  typedef TPanel inherited;
  private:
  WideString FFileName;
  void* FFileHandle;
  __int64 FFileSize;
  bool FFileOK;
  TATUnicodeFormat FFileUnicodeFmt;
  TATFileSource FFileSourceType;
  TATStreamSearch* FSearch;
  bool FSearchStarted;
  TStream* FStream;
  char* FBuffer;
  int FBufferMaxOffset;
  int FBufferAllocSize;
  bool FLinesShow;
  int FLinesStep;
  int FLinesBufSize;
  PATBinHexIntArray FLinesData;
  int FLinesNum;
  int FLinesCount;
  bool FLinesExtUse;
  AnsiString FLinesExtList;
  Graphics::TBitmap* FBitmap;
  TTimer* FTimerAutoScroll;
  TTimer* FTimerNiceScroll;
  TObject* FStrings;
  TPopupMenu* FMenu;
  TMenuItem* FMenuItemCopy;
  TMenuItem* FMenuItemCopyHex;
  TMenuItem* FMenuItemCopyLink;
  TMenuItem* FMenuItemSelectLine;
  TMenuItem* FMenuItemSelectAll;
  TMenuItem* FMenuItemEncMenu;
  TMenuItem* FMenuItemSep1;
  TMenuItem* FMenuItemSep2;
  TPopupMenu* FMenuCodepages;
  TPopupMenu* FMenuCodepagesUn;
  TATFileNotification* FNotif;
  TATFileNotificationSimple* FNotif2;
  bool FAutoReload;
  bool FAutoReloadBeep;
  bool FAutoReloadFollowTail;
  bool FAutoReloadSimple;
  int FAutoReloadSimpleTime;
  int FLockCount;
  __int64 FBufferPos;
  __int64 FViewPos; //Position of view area (bytes)
  bool FViewAtEnd; //Shows if we are at the end of file, after redraw
  __int64 FViewPageSize; //Page size (number of bytes on screen), after redraw
  int FHViewPos; //Horizontal scroll position (px)
  int FHViewWidth; //Horizontal width of text on screen, after redraw
  __int64 FSelStart;
  __int64 FSelLength;
  TATBinHexMode FMode;
  TATEncoding FEncoding;
  TATUrlArray FUrlArray;
  bool FUrlShow;
  int FTextWidth;
  int FTextWidthHex;
  int FTextWidthUHex;
  bool FTextWidthFit;
  bool FTextWidthFitHex;
  bool FTextWidthFitUHex;
  bool FTextWrap;
  bool FTextNonPrintable;
  bool FTextOemSpecial;
  bool FTextGutter;
  int FTextGutterWidth;
  TColor FTextColorHex;
  TColor FTextColorHex2;
  TColor FTextColorHexBack;
  TColor FTextColorLines;
  TColor FTextColorError;
  TColor FTextColorGutter;
  TColor FTextColorURL;
  int FSearchIndentVert;
  int FSearchIndentHorz;
  int FTabSize;
  TATPopupCommands FPopupCommands;
  bool FEnabled2;
  int FMaxLength;
  int FMaxLengths [ 5 /*# TATBinHexMode */ ];
  int FMaxClipboardDataSizeMb;
  TFont* FFontOEM;
  TFont* FFontFooter;
  TFont* FFontGutter;
  int FHexOffsetLen;
  int FFontHeight;
  char FFontFirstChar;
  int FFontWidthDigits;
  bool FFontMonospaced;
  bool FMouseDown;
  __int64 FMouseStart;
  __int64 FMouseStartShift;
  __int64 FMouseStartDbl;
  bool FMouseDblClick;
  bool FMouseTriClick;
  DWORD FMouseTriTime;
  TPoint FMousePopupPos;
  TATMouseRelativePosition FMouseRelativePos;
  bool FMouseNiceScroll;
  TPoint FMouseNiceScrollPos;
  int FClientHeight;
  TNotifyEvent FOnSelectionChange;
  TNotifyEvent FOnOptionsChange;
  TNotifyEvent FOnScroll;
  TATBinHexDrawLine FOnDrawLine;
  TATBinHexDrawLine2 FOnDrawLine2;
  TATBinHexClickURL FOnClickURL;
  TNotifyEvent FOnFileReload;
  long double FMarginLeft;
  long double FMarginTop;
  long double FMarginRight;
  long double FMarginBottom;
  bool FPrintFooter;
  void __fastcall AllocBuffer( );
  bool __fastcall SourceAssigned( );
  bool __fastcall ReadSource( const __int64 APos, void* ABuffer, DWORD ABufferSize, DWORD& AReadSize );
  void __fastcall ReadBuffer( const __int64 APos = - 1 );
  void __fastcall InitData( );
  void __fastcall FreeData( );
  bool __fastcall LoadFile( bool ANewFile );
  bool __fastcall LoadStream( );
  __int64 __fastcall PosBefore( const __int64 APos, TATLineType ALineType, TATDirection ADir );
  void __fastcall ReadUnicodeFmt( );
  void __fastcall HideScrollbars( );
  void __fastcall UpdateVertScrollbar( );
  void __fastcall UpdateHorzScrollbar( );
  void __fastcall SetMode( TATBinHexMode AMode );
  void __fastcall SetTextEncoding( TATEncoding AValue );
  void __fastcall SetTextWidthTo( int AValue, int& AField );
  void __fastcall SetTextWidthHexTo( int AValue, int& AField );
  void __fastcall SetTextWidthUHexTo( int AValue, int& AField );
  void __fastcall SetTextWidth( int AValue );
  void __fastcall SetTextWidthHex( int AValue );
  void __fastcall SetTextWidthUHex( int AValue );
  void __fastcall SetTextWidthFit( bool AValue );
  void __fastcall SetTextWidthFitHex( bool AValue );
  void __fastcall SetTextWidthFitUHex( bool AValue );
  void __fastcall SetTextWrap( bool AValue );
  void __fastcall SetTextNonPrintable( bool AValue );
  void __fastcall SetTextUrlHilight( bool AValue );
  void __fastcall SetSearchIndentVert( int AValue );
  void __fastcall SetSearchIndentHorz( int AValue );
  void __fastcall SetFontOEM( TFont* AValue );
  void __fastcall SetFontFooter( TFont* AValue );
  void __fastcall SetFontGutter( TFont* AValue );
  void __fastcall SetLinesBufSize( int AValue );
  void __fastcall SetLinesCount( int AValue );
  void __fastcall SetLinesStep( int AValue );
  void __fastcall InitHexOffsetLen( );
  void __fastcall MsgReadError( );
  bool __fastcall MsgReadRetry( );
  void __fastcall MsgOpenError( );
  int __fastcall DrawOffsetX( );
  int __fastcall DrawOffsetY( );
  void __fastcall SetTextGutter( bool AValue );
  int __fastcall LinesNum( Graphics::TBitmap* ABitmap = NULL );
  int __fastcall ColsNumFit( Graphics::TBitmap* ABitmap = NULL );
  int __fastcall ColsNumHexFit( Graphics::TBitmap* ABitmap = NULL );
  int __fastcall ColsNumUHexFit( Graphics::TBitmap* ABitmap = NULL );
  int __fastcall ColsNum( Graphics::TBitmap* ABitmap = NULL );
  bool __fastcall PosBad( const __int64 APos );
  __int64 __fastcall PosMax( );
  __int64 __fastcall PosLast( );
  void __fastcall PosFixCRLF( __int64& APos );
  void __fastcall PosAt( const __int64 APos, bool ARedraw = true );
  void __fastcall PosDec( const __int64 n );
  void __fastcall PosInc( const __int64 n );
  void __fastcall PosLineUp( int ALines = 1 ) /*# overload */;
  void __fastcall PosLineDown( int ALines = 1 ) /*# overload */;
  void __fastcall PosLineUp( bool AViewAtEnd, int ALines ) /*# overload */;
  void __fastcall PosLineDown( bool AViewAtEnd, int ALines ) /*# overload */;
  void __fastcall PosPageUp( );
  void __fastcall PosPageDown( );
  void __fastcall PosBegin( );
  void __fastcall PosEndTry( );
  void __fastcall PosEnd( );
  void __fastcall HPosAt( int APos, bool ARedraw = true );
  void __fastcall HPosInc( int n );
  void __fastcall HPosDec( int n );
  void __fastcall HPosBegin( );
  void __fastcall HPosEnd( );
  void __fastcall HPosLeft( );
  void __fastcall HPosRight( );
  void __fastcall HPosPageLeft( );
  void __fastcall HPosPageRight( );
  int __fastcall HPosWidth( );
  int __fastcall HPosMax( );
  bool __fastcall LineWithCR( const __int64 APos, const WideString ALine );
  bool __fastcall LineWithGutterDot( const __int64 APos );
  TATBinHexOutputOptions __fastcall OutputOptions( bool AShowCR = false );
  int __fastcall GetPosPercent( );
  void __fastcall SetPosPercent( int APos );
  int __fastcall GetPosLine( );
  void __fastcall SetPosLine( int ALine );
  __int64 __fastcall GetPosOffset( );
  void __fastcall SetPosOffset( const __int64 APos );
  void __fastcall MouseWheelUp( TObject* Sender, TShiftState Shift, const TPoint& MousePos, bool& Handled );
  void __fastcall MouseWheelDown( TObject* Sender, TShiftState Shift, const TPoint& MousePos, bool& Handled );
  __int64 __fastcall MousePosition( int AX, int AY, bool AStrict = false );
  void __fastcall MouseMoveAction( int AX, int AY );
  void __fastcall TimerAutoScrollTimer( TObject* Sender );
  void __fastcall TimerNiceScrollTimer( TObject* Sender );
  void __fastcall MenuItemCopyClick( TObject* Sender );
  void __fastcall MenuItemCopyHexClick( TObject* Sender );
  void __fastcall MenuItemCopyLinkClick( TObject* Sender );
  void __fastcall MenuItemSelectLineClick( TObject* Sender );
  void __fastcall MenuItemSelectAllClick( TObject* Sender );
  void __fastcall MenuItemEncMenuClick( TObject* Sender );
  void __fastcall UpdateMenu( TObject* Sender );
  void __fastcall SetTextPopupCaption( TATPopupCommand AIndex, const AnsiString AValue );
  void __fastcall SetTabSize( int AValue );
  void __fastcall InitURLs( );
  void __fastcall FindURLs( DWORD ABufSize );
  AnsiString __fastcall PosURL( const __int64 APos );
  bool __fastcall IsPosURL( const __int64 APos );
  int __fastcall FindLineLength( const __int64 AStartPos, TATDirection ADir, WideString& ALine );
  __int64 __fastcall FindLinePos( const __int64 AStartPos, TATDirection ADir, WideString& ALine, bool APassiveMove = false );
  void __fastcall PosNextLineFrom( const __int64 AStartPos, int ALinesNum, TATDirection ADir, bool APassiveMove = false, bool ARedraw = true );
  void __fastcall PosNextLine( int ALinesNum, TATDirection ADir, bool AViewAtEnd );
  wchar_t __fastcall GetChar( const __int64 ACharPos );
  WideString __fastcall GetHex( const __int64 ACharPos );
  WideString __fastcall DecodeString( const WideString S );
  int __fastcall CharSize( );
  bool __fastcall IsFileEmpty( );
  bool __fastcall IsModeVariable( );
  bool __fastcall IsModeUnicode( );
  bool __fastcall IsUnicodeBE( );
  void __fastcall NormalizePos( __int64& APos );
  __int64 __fastcall NormalizedPos( const __int64 APos );
  void __fastcall NextPos( __int64& APos, TATDirection ADir, int AChars = 1 );
  void __fastcall SelectLineAtPos( const __int64 APos, TATLineType ALineType );
  void __fastcall ContextPopup( TObject* Sender, const TPoint& MousePos, bool& Handled );
  TFont* __fastcall ActiveFont( );
  bool __fastcall ActiveLinesShow( );
  void __fastcall DrawGutterTo( Graphics::TBitmap* ABitmap );
  void __fastcall DrawEmptyTo( Graphics::TBitmap* ABitmap, int APageWidth, int APageHeight, bool APrintMode );
  void __fastcall NotifChanged( TObject* Sender );
  void __fastcall DoFileReload( );
  void __fastcall SetAutoReload( bool AValue );
  void __fastcall Lock( );
  void __fastcall Unlock( );
  bool __fastcall Locked( );
  AnsiString __fastcall GetSelTextRaw( int AMaxSize = 0 );
  AnsiString __fastcall GetSelText( );
  AnsiString __fastcall GetSelTextShort( );
  WideString __fastcall GetSelTextW( );
  WideString __fastcall GetSelTextShortW( );
  void __fastcall DoSelectionChange( );
  int __fastcall GetMaxLengths( TATBinHexMode AIndex );
  void __fastcall SetMaxLengths( TATBinHexMode AIndex, int AValue );
  void __fastcall SetMaxClipboardDataSizeMb( int AValue );
  void __fastcall SetEnabled2( bool AValue );
  void __fastcall SetMouseNiceScroll( bool AValue );
  void __fastcall DrawNiceScroll( );
  __property bool MouseNiceScroll = { read = FMouseNiceScroll, write = SetMouseNiceScroll };
  void __fastcall ExitProc( TObject* Sender );
  void __fastcall EncodingMenuItemClick( TObject* Sender );
  void __fastcall EncodingMenuUnItemClick( TObject* Sender );
  AnsiString __fastcall GetTextEncodingName( );
  void __fastcall SetFileUnicodeFmt( TATUnicodeFormat AValue );
  void __fastcall SetTextOemSpecial( bool AValue );
  bool __fastcall CountLines( int ABufSize );
  bool __fastcall GetLineNumberOffset( int ALine, bool AFindLine, int& ACurrentLine, __int64& aoffset );
  int __fastcall FindLineNum( const __int64 aoffset );
  WideString __fastcall StringAtPos( const __int64 APos );
  TATStreamSearchProgress __fastcall GetOnSearchProgress( );
  void __fastcall SetOnSearchProgress( TATStreamSearchProgress AValue );
  __int64 __fastcall GetSearchResultStart( );
  __int64 __fastcall GetSearchResultLength( );
  WideString __fastcall GetSearchString( );
  void __fastcall DoOptionsChange( );
  void __fastcall DoScroll( );
  void __fastcall DoDrawLine( TCanvas* ACanvas, const WideString AStr, const __int64 APos, const TRect& ARect, const TPoint& ATextPnt, bool& ADone );
  void __fastcall DoDrawLine2( TCanvas* ACanvas, const WideString AStr, const TPoint& APnt, const TATBinHexOutputOptions& AOptions );
  void __fastcall DoClickURL( const __int64 AMousePos );
  protected:
  DYNAMIC void __fastcall DblClick( );
  DYNAMIC void __fastcall Resize( );
  virtual void __fastcall Paint( );
  MESSAGE void __fastcall WMGetDlgCode( TMessage& Message ) /*# WM_GETDLGCODE */;
  MESSAGE void __fastcall WMEraseBkgnd( TMessage& Message ) /*# WM_ERASEBKGND */;
  MESSAGE void __fastcall WMVScroll( TWMVScroll& Message ) /*# WM_VSCROLL */;
  MESSAGE void __fastcall WMHScroll( TWMHScroll& Message ) /*# WM_HSCROLL */;
  DYNAMIC void __fastcall KeyDown( WORD& key, TShiftState Shift );
  DYNAMIC void __fastcall MouseDown( TMouseButton Button, TShiftState Shift, int x, int y );
  DYNAMIC void __fastcall MouseMove( TShiftState Shift, int x, int y );
  DYNAMIC void __fastcall MouseUp( TMouseButton Button, TShiftState Shift, int x, int y );
  virtual void __fastcall CreateParams( TCreateParams& Params );
  virtual void __fastcall SetEnabled( bool AValue );
  public:
  __fastcall TATBinHex( TComponent* AOwner );
  __fastcall virtual ~TATBinHex( );
  bool __fastcall Open( const WideString AFileName, bool ARedraw = true );
  bool __fastcall OpenStream( TStream* AStream, bool ARedraw = true );
  void __fastcall Reload( );
  void __fastcall Redraw( );
  bool __fastcall FindFirst( const WideString AText, TATStreamSearchOptions AOptions );
  bool __fastcall FindNext( bool AFindPrevious = false );
  __property __int64 SearchResultStart = { read = GetSearchResultStart };
  __property __int64 SearchResultLength = { read = GetSearchResultLength };
  __property bool SearchStarted = { read = FSearchStarted };
  __property WideString SearchString = { read = GetSearchString };
  bool __fastcall IncreaseFontSize( bool AIncrement );
  void __fastcall CopyToClipboard( bool AAsHex = false );
  __property __int64 SelStart = { read = FSelStart };
  __property __int64 SelLength = { read = FSelLength };
  __property AnsiString SelText = { read = GetSelText };
  __property AnsiString SelTextShort = { read = GetSelTextShort };
  __property WideString SelTextW = { read = GetSelTextW };
  __property WideString SelTextShortW = { read = GetSelTextShortW };
  void __fastcall SetSelection( const __int64 AStart, const __int64 ALength, bool AScroll, bool AFireEvent = true, bool ARedraw = true );
  void __fastcall Scroll( const __int64 APos, int AIndentVert, int AIndentHorz, bool ARedraw = true );
  void __fastcall SelectAll( );
  void __fastcall SelectNone( bool AFireEvent = true );
  void __fastcall DrawTo( Graphics::TBitmap* ABitmap, int APageWidth, int APageHeight, TObject* AStringsObject, bool APrintMode, const __int64 AFinalPos, int& ATextWidth, int& ATextWidthHex, int& ATextWidthUHex, __int64& AViewPageSize, bool& AViewAtEnd );
  AnsiString __fastcall PrinterCaption( );
  WideString __fastcall PrinterFooter( int APageNumber );
  void __fastcall PrintPreview( );
  void __fastcall PrintTo( TCanvas* ACanvas, //ACanvas may be assigned only for Print Preview
  int APageWidth, int APageHeight, TPrintRange APrintRange, int AFromPage, int AToPage );
  void __fastcall Print( TPrintRange APrintRange, int AFromPage = 1, int AToPage = MaxInt, int ACopies = 1, const AnsiString ACaption = "" );
  TRect __fastcall MarginsRectPx( int ATargetWidth, int ATargetHeight, int ATargetPPIX, int ATargetPPIY );
  TRect __fastcall MarginsRectRealPx( );
  __property long double MarginLeft = { read = FMarginLeft, write = FMarginLeft };
  __property long double MarginTop = { read = FMarginTop, write = FMarginTop };
  __property long double MarginRight = { read = FMarginRight, write = FMarginRight };
  __property long double MarginBottom = { read = FMarginBottom, write = FMarginBottom };
  __property bool PrintFooter = { read = FPrintFooter, write = FPrintFooter };
  __property int PosPercent = { read = GetPosPercent, write = SetPosPercent };
  __property __int64 PosOffset = { read = GetPosOffset, write = SetPosOffset };
  __property int PosLine = { read = GetPosLine, write = SetPosLine };
  __property AnsiString TextPopupCaption [ TATPopupCommand AIndex ] = { write = SetTextPopupCaption };
  __property int MaxLengths [ TATBinHexMode AIndex ] = { read = GetMaxLengths, write = SetMaxLengths };
  __property int MaxClipboardDataSizeMb = { read = FMaxClipboardDataSizeMb, write = SetMaxClipboardDataSizeMb };
  __property WideString Filename = { read = FFileName };
  __property __int64 FileSize = { read = FFileSize };
  __property bool FileReadOK = { read = FFileOK };
  __property TATUnicodeFormat FileUnicodeFormat = { read = FFileUnicodeFmt, write = SetFileUnicodeFmt };
  __property AnsiString TextEncodingName = { read = GetTextEncodingName };
  void __fastcall TextEncodingsMenu( int AX, int AY );

    //Enabled2 is the same as Enabled, but also enables control redrawing:
    //we need to disable it during printing.
  __property bool Enabled2 = { read = FEnabled2, write = SetEnabled2 };
  public: inline __fastcall TATBinHex( HWND ParentWindow ) : inherited( ParentWindow ) {}
  __published:
  __property TFont* FontOEM = { read = FFontOEM, write = SetFontOEM };
  __property TFont* FontFooter = { read = FFontFooter, write = SetFontFooter };
  __property TFont* FontGutter = { read = FFontGutter, write = SetFontGutter };
  __property TATBinHexMode Mode = { read = FMode, write = SetMode, default = vbmodeText };
  __property TATEncoding TextEncoding = { read = FEncoding, write = SetTextEncoding, default = vEncANSI };
  __property int TextWidth = { read = FTextWidth, write = SetTextWidth, default = 80 };
  __property int TextWidthHex = { read = FTextWidthHex, write = SetTextWidthHex, default = 16 };
  __property int TextWidthUHex = { read = FTextWidthUHex, write = SetTextWidthUHex, default = 8 };
  __property bool TextWidthFit = { read = FTextWidthFit, write = SetTextWidthFit, default = false };
  __property bool TextWidthFitHex = { read = FTextWidthFitHex, write = SetTextWidthFitHex, default = false };
  __property bool TextWidthFitUHex = { read = FTextWidthFitUHex, write = SetTextWidthFitUHex, default = false };
  __property bool TextWrap = { read = FTextWrap, write = SetTextWrap, default = false };
  __property bool TextNonPrintable = { read = FTextNonPrintable, write = SetTextNonPrintable, default = false };
  __property bool TextOemSpecial = { read = FTextOemSpecial, write = SetTextOemSpecial, default = false };
  __property bool TextUrlHilight = { read = FUrlShow, write = SetTextUrlHilight, default = true };
  __property bool TextGutter = { read = FTextGutter, write = SetTextGutter, default = false };
  __property bool TextGutterLines = { read = FLinesShow, write = FLinesShow, default = true };
  __property int TextGutterLinesStep = { read = FLinesStep, write = SetLinesStep, default = 5 };
  __property int TextGutterLinesCount = { read = FLinesCount, write = SetLinesCount, stored = false };
  __property int TextGutterLinesBufSize = { read = FLinesBufSize, write = SetLinesBufSize, stored = false };
  __property bool TextGutterLinesExtUse = { read = FLinesExtUse, write = FLinesExtUse, default = false };
  __property AnsiString TextGutterLinesExtList = { read = FLinesExtList, write = FLinesExtList };
  __property TColor TextColorHex = { read = FTextColorHex, write = FTextColorHex, default = /*# clNavy */ ((TColor)0x800000) };
  __property TColor TextColorHex2 = { read = FTextColorHex2, write = FTextColorHex2, default = /*# clBlue */ ((TColor)0xFF0000) };
  __property TColor TextColorHexBack = { read = FTextColorHexBack, write = FTextColorHexBack, default = /*# cATBinHexBkColor */ 15527148 };
  __property TColor TextColorLines = { read = FTextColorLines, write = FTextColorLines, default = /*# clGray */ ((TColor)0x808080) };
  __property TColor TextColorError = { read = FTextColorError, write = FTextColorError, default = /*# clRed */ ((TColor)0x0000FF) };
  __property TColor TextColorGutter = { read = FTextColorGutter, write = FTextColorGutter, default = /*# clLtGray */ ((TColor)0xC0C0C0) };
  __property TColor TextColorURL = { read = FTextColorURL, write = FTextColorURL, default = /*# clBlue */ ((TColor)0xFF0000) };
  __property int TextSearchIndentVert = { read = FSearchIndentVert, write = SetSearchIndentVert, default = 5 };
  __property int TextSearchIndentHorz = { read = FSearchIndentHorz, write = SetSearchIndentHorz, default = 5 };
  __property int TextTabSize = { read = FTabSize, write = SetTabSize, default = 8 };
  __property TATPopupCommands TextPopupCommands = { read = FPopupCommands, write = FPopupCommands, default = /*# cATBinHexCommandSet */ 63 };
  __property bool AutoReload = { read = FAutoReload, write = SetAutoReload, default = false };
  __property bool AutoReloadBeep = { read = FAutoReloadBeep, write = FAutoReloadBeep, default = false };
  __property bool AutoReloadFollowTail = { read = FAutoReloadFollowTail, write = FAutoReloadFollowTail, default = true };
  __property bool AutoReloadSimple = { read = FAutoReloadSimple, write = FAutoReloadSimple, default = false };
  __property int AutoReloadSimpleTime = { read = FAutoReloadSimpleTime, write = FAutoReloadSimpleTime, default = 1000 };
  __property TNotifyEvent OnFileReload = { read = FOnFileReload, write = FOnFileReload };
  __property TATStreamSearchProgress OnSearchProgress = { read = GetOnSearchProgress, write = SetOnSearchProgress };
  __property TNotifyEvent OnSelectionChange = { read = FOnSelectionChange, write = FOnSelectionChange };
  __property TNotifyEvent OnOptionsChange = { read = FOnOptionsChange, write = FOnOptionsChange };
  __property TNotifyEvent OnScroll = { read = FOnScroll, write = FOnScroll };
  __property TATBinHexDrawLine OnDrawLine = { read = FOnDrawLine, write = FOnDrawLine };
  __property TATBinHexDrawLine2 OnDrawLine2 = { read = FOnDrawLine2, write = FOnDrawLine2 };
  __property TATBinHexClickURL OnClickURL = { read = FOnClickURL, write = FOnClickURL };

  private:
  BEGIN_MESSAGE_MAP
    VCL_MESSAGE_HANDLER(WM_GETDLGCODE, TMessage, WMGetDlgCode)
    VCL_MESSAGE_HANDLER(WM_ERASEBKGND, TMessage, WMEraseBkgnd)
    VCL_MESSAGE_HANDLER(WM_VSCROLL, TWMVScroll, WMVScroll)
    VCL_MESSAGE_HANDLER(WM_HSCROLL, TWMHScroll, WMHScroll)
  END_MESSAGE_MAP( TPanel )


  private: // originally nested
  void __fastcall SelectLine( const WideString ALine, int AX, int AY, const __int64 AFilePos, bool ASelectAll, Graphics::TBitmap*& ABitmap, int& APageWidth, int& APageHeight, TObject*& AStringsObject, bool& APrintMode, const __int64& AFinalPos, int& ATextWidth, int& ATextWidthHex, int& ATextWidthUHex, __int64& AViewPageSize, bool& AViewAtEnd, TATStringExtent& DX, TColor& AColorBack, TColor& AColorBackHex, TColor& AColorText, TColor& AColorTextHex1, TColor& AColorTextHex2, TColor& AColorLines, TColor& AColorError, TColor& AColorURL );
  void __fastcall HilightLine( const WideString ALine, int AX, int AY, const __int64 AFilePos, Graphics::TBitmap*& ABitmap, int& APageWidth, int& APageHeight, TObject*& AStringsObject, bool& APrintMode, const __int64& AFinalPos, int& ATextWidth, int& ATextWidthHex, int& ATextWidthUHex, __int64& AViewPageSize, bool& AViewAtEnd, TATStringExtent& DX, TColor& AColorBack, TColor& AColorBackHex, TColor& AColorText, TColor& AColorTextHex1, TColor& AColorTextHex2, TColor& AColorLines, TColor& AColorError, TColor& AColorURL );
  TColor __fastcall ActiveColor( TColor AColor, Graphics::TBitmap*& ABitmap, int& APageWidth, int& APageHeight, TObject*& AStringsObject, bool& APrintMode, const __int64& AFinalPos, int& ATextWidth, int& ATextWidthHex, int& ATextWidthUHex, __int64& AViewPageSize, bool& AViewAtEnd, TATStringExtent& DX, TColor& AColorBack, TColor& AColorBackHex, TColor& AColorText, TColor& AColorTextHex1, TColor& AColorTextHex2, TColor& AColorLines, TColor& AColorError, TColor& AColorURL );
  __int64 __fastcall PrevPos( const __int64 APos, const __int64& AStartPos, TATDirection& ADir, WideString& ALine, bool& APassiveMove, __int64& xResult );
  bool __fastcall PrevLine( __int64& APos, const __int64& AStartPos, TATDirection& ADir, WideString& ALine, bool& APassiveMove, __int64& xResult );
  void __fastcall DrawFooter( TCanvas* ACanvas, const TRect& ARect, int& APageWidth, int& APageHeight, TPrintRange& APrintRange, int& AFromPage, int& AToPage, int& APageNumber );
  void __fastcall DrawRect( TCanvas* ACanvas, const TRect& ARect, int& APageWidth, int& APageHeight, TPrintRange& APrintRange, int& AFromPage, int& AToPage, int& APageNumber );
  int __fastcall FooterHeightSrc( TCanvas*& ACanvas, int& APageWidth, int& APageHeight, TPrintRange& APrintRange, int& AFromPage, int& AToPage, int& APageNumber );
  char __fastcall AChar( const __int64 APos, int& ABufSize, bool& xResult, char*& Buf, __int64& aoffset );
  void __fastcall AInc( int& ABufSize, bool& result, char*& Buf, __int64& aoffset );
};


bool __fastcall TextIncreaseFontSize( TFont* AFont, TCanvas* ACanvas, bool AIncrement );
int __fastcall STextWidth( TCanvas* ACanvas, const WideString S );
void __fastcall STextOut( TCanvas* ACanvas, int x, int y, const WideString S );
void __fastcall Register( );

#endif //  ATBinHexH