#include <vcl.h>
#pragma hdrstop

/*
    Copyright of the basic file version
    -----------------------------------

    Explanation: in the following license "library" means
    the following files: 
    
    System.pas / d2c_system.pas
    System.h
    d2c_sysconst.h
    d2c_syscurr.h d2c_syscurr.cpp
    d2c_sysdate.h d2c_sysdate.cpp
    d2c_sysexcept.h d2c_sysexcept.cpp
    d2c_sysfile.h, d2c_sysfile.cpp
    d2c_sysmath.h d2c_sysmath.cpp
    d2c_sysstring.h d2c_sysstring.cpp
    d2c_systypes.h
    d2c_varrec.h
    d2c_smallstring.h
    Windows.pas
    Sysutils.pas
    Sysutils.h Sysutils.cpp

    AS THEY ARE CONTAINED IN THE FREE TRIAL VERSION OF Delphi2Cpp.

    This library is derived from the FreePascal library:

    http://www.freepascal.org/

    FreePascal is published under the terms of GNU Lesser General
    Public License and the same terms apply to this library.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with d2c_sysfile.h/cpp; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



    Copyright of the extended file version
    --------------------------------------


    The terms of the license above don't apply to extended versions
    of these files, which are distributed with commercial versions of 
    Delphi2Cpp. Individual licenses are applied to them. 
    The library doesn't depend on the commercial extensions and the 
    the commercial extensions only originates from the author
    Dr. Detlef Meyer-Eltz or might use code which has no copyright restrictions

    Copyright (C) <2011>  <Dr. Detlef Meyer-Eltz>

    The extended version of this file is authorized for unlimited use in any 
    Delphi2Cpp project.
	     
	  http://www.texttransformer.com/Delphi2Cpp_en.html

*/


#include "d2c_system.h"
#include "d2c_sysmath.h"

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <stdexcept>
#include <iomanip>

#if defined(windows) && defined (_WIDESTRING)
#include <ShellAPI.h> // CommandLineToArgvW
#endif

using namespace std;



// #define _CRT_NON_CONFORMING_SWPRINTFS 1




namespace d2c_system
{



void FillChar( void* X, int Count, unsignedchar Value )
{
  memset(X, Value, Count);
}

void FillChar( char* X, int Count, unsignedchar Value )
{
  memset(X, Value, Count);
}

void FillChar( wchar_t* X, int Count, unsignedchar Value )
{
  wchar_t* p = X;
  if(Count <= 0)
     return;
  while(Count--)
    *p++ = (wchar_t) Value;
}

void FillChar( AnsiString& X, int Count, unsignedchar Value )
{
  if(Count <= 0)
     return;
  for(int i = 1; i <= Count; i++)
   X[i] = Value;
}

void FillChar( WideString& X, int Count, unsignedchar Value )
{
  if(Count <= 0)
     return;
  for(int i = 1; i <= Count; i++)
   X[i] = Value;
}

void Str(long double xd, AnsiString& xs)
{
  ostringstream oss;
  oss << scientific << uppercase << setprecision(14) << xd;
  xs = oss.str().c_str();
}

void Str(double xd, AnsiString& xs)
{
  ostringstream oss;
  oss << scientific << uppercase << setprecision(14) << xd;
  xs = oss.str().c_str();
}

void Str(int xi, int xiMinWidth, AnsiString& xs)
{
  ostringstream oss;
  oss << scientific << uppercase << setw(xiMinWidth) << setprecision(14) << xi;
  xs = oss.str().c_str();
}

void Str(double xd, int xiMinWidth, AnsiString& xs)
{
  ostringstream oss;
  oss << scientific << uppercase << setw(xiMinWidth) << setprecision(14)  << xd;
  xs = oss.str().c_str();
}

void Str(long double xd, int xiMinWidth, AnsiString& xs)
{
  ostringstream oss;
  oss << scientific << uppercase << setw(xiMinWidth) << setprecision(14)  << xd;
  xs = oss.str().c_str();
}

void Str(double xd, int xiMinWidth, int xiDecPlaces, AnsiString& xs)
{
  ostringstream oss;
  if(xiMinWidth == 0)
  {
    if(xiDecPlaces == 0)
     oss << fixed << setprecision(xiDecPlaces) << xd;
    else
     oss << fixed << showpoint << setprecision(xiDecPlaces) << xd;
  }
  else
    oss << scientific << uppercase << setw(xiMinWidth) << setprecision(xiDecPlaces) << xd;
  xs = oss.str().c_str();
}

void Str(long double xd, int xiMinWidth, int xiDecPlaces, AnsiString& xs)
{
  ostringstream oss;
  if(xiMinWidth == 0)
  {
    if(xiDecPlaces == 0)
     oss << fixed << setprecision(xiDecPlaces) << xd;
    else
     oss << fixed << showpoint << setprecision(xiDecPlaces) << xd;
  }
  else
    oss << scientific << uppercase << setw(xiMinWidth) << setprecision(xiDecPlaces) << xd;
  xs = oss.str().c_str();
}

void Str(const Currency& xcr, int xiMinWidth, int xiDecPlaces, AnsiString& xs)
{
  ostringstream oss;
  if(xiMinWidth == 0)
  {
    if(xiDecPlaces == 0)
     oss << fixed << setprecision(xiDecPlaces) << (double) xcr;
    else
     oss << fixed << showpoint << setprecision(xiDecPlaces) << (double) xcr;
  }
  else
    oss << setw(xiMinWidth) << setprecision(xiDecPlaces) << (double) xcr;
  xs = oss.str().c_str();
}


void Str(long double xd, WideString& xs)
{
  wostringstream oss;
  oss << scientific << uppercase << setprecision(14) << xd;
  xs = (wchar_t*) oss.str().c_str();
}

void Str(double xd, WideString& xs)
{
  wostringstream oss;
  oss << scientific << uppercase << setprecision(14) << xd;
  xs = (wchar_t*) oss.str().c_str();
}


void Str(double xd, int xiMinWidth, WideString& xs)
{
  wostringstream oss;
  oss << scientific << uppercase << setw(xiMinWidth) << setprecision(14)  << xd;
  xs = (wchar_t*) oss.str().c_str();
}

void Str(int xi, int xiMinWidth, WideString& xs)
{
  wostringstream oss;
  oss << scientific << uppercase << setw(xiMinWidth) << setprecision(14)  << xi;
  xs = (wchar_t*) oss.str().c_str();
}


void Str(long double xd, int xiMinWidth, WideString& xs)
{
  wostringstream oss;
  oss << scientific << uppercase << setw(xiMinWidth) << setprecision(14) << xd;
  xs = (wchar_t*) oss.str().c_str();
}

void Str(double xd, int xiMinWidth, int xiDecPlaces, WideString& xs)
{
  wostringstream oss;
  if(xiMinWidth == 0)
  {
    if(xiDecPlaces == 0)
     oss << fixed << setprecision(xiDecPlaces) << xd;
    else
     oss << fixed << showpoint << setprecision(xiDecPlaces) << xd;
  }
  else
    oss << scientific << uppercase << setw(xiMinWidth) << setprecision(xiDecPlaces) << xd;
  xs = (wchar_t*) oss.str().c_str();
}

void Str(long double xd, int xiMinWidth, int xiDecPlaces, WideString& xs)
{
  wostringstream oss;
  if(xiMinWidth == 0)
  {
    if(xiDecPlaces == 0)
     oss << fixed << setprecision(xiDecPlaces) << xd;
    else
     oss << fixed << showpoint << setprecision(xiDecPlaces) << xd;
  }
  else
    oss << scientific << uppercase << setw(xiMinWidth) << setprecision(xiDecPlaces) << xd;
  xs = (wchar_t*) oss.str().c_str();
}

void Str(const Currency& xcr, int xiMinWidth, int xiDecPlaces, WideString& xs)
{
  wostringstream oss;
  if(xiMinWidth == 0)
  {
    if(xiDecPlaces == 0)
     oss << fixed << setprecision(xiDecPlaces) << (double) xcr;
    else
     oss << fixed << showpoint << setprecision(xiDecPlaces) << (double) xcr;
  }
  else
    oss << setw(xiMinWidth) << setprecision(xiDecPlaces) << (double) xcr;
  xs = (wchar_t*) oss.str().c_str();
}


WORD Swap( WORD X )
{
  return ((WORD) ( ( X >> 8 ) | ( X << 8 ) ) );
}

int Swap( int X )
{
  return ( ( X & 0xFFFF ) << 16 ) + ( X >> 16 );
}

unsignedint Swap( unsignedint X )
{
  return ( ( X & 0xFFFF ) << 16 ) + ( X >> 16 );
}

int64_t Swap( int64_t X )
{
  return ( ( X & 0xFFFFFFFF ) << 32 ) + ( X >> 32 );
}

/////////

void __fastcall Assert( bool expr, const String Msg )
{
  if ( ! expr )
	throw Exception(Msg);
}

long double __fastcall Sqr( long double x )
{
  return x * x;
}

int __fastcall Inc(int& xi)
{
  return ++xi;
}

void __fastcall SetString( AnsiString& S, char* Buffer, int Len )
{
  S = AnsiString(Buffer, Len);
}

void __fastcall SetString( WideString& S, wchar_t* Buffer, int Len )
{
  S = WideString(Buffer, Len);
}


}  // namespace System
