//{$define STREAMS} //See Readme.txt

unit UFormView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls,
  {$ifdef STREAMS}TntClasses,{$endif}
  ATBinHex, Menus;

type
  TFormView = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Open1: TMenuItem;
    OpenDialog1: TOpenDialog;
    GroupBox2: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    chkModeText: TRadioButton;
    chkModeBinary: TRadioButton;
    chkModeHex: TRadioButton;
    chkModeUnicode: TRadioButton;
    chkModeUHex: TRadioButton;
    chkOEM: TCheckBox;
    chkWordWrap: TCheckBox;
    chkEnabled: TCheckBox;
    chkNonPrint: TCheckBox;
    procedure chkModeBinaryClick(Sender: TObject);
    procedure chkModeHexClick(Sender: TObject);
    procedure chkModeUnicodeClick(Sender: TObject);
    procedure chkModeTextClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure chkWordWrapClick(Sender: TObject);
    procedure chkOEMClick(Sender: TObject);
    procedure chkEnabledClick(Sender: TObject);
    procedure chkModeUHexClick(Sender: TObject);
    procedure chkNonPrintClick(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure ViewerDrawLine(ASender: TObject; ACanvas: TCanvas; const AStr,
      AStrAll: WideString; const ARect: TRect; const ATextPnt: TPoint;
      var ADone: Boolean);
  private
    { Private declarations }
    {$ifdef STREAMS}
    FStream: TTntFileStream;
    {$endif}
    FLoad: boolean;
  public
    { Public declarations }
    procedure UpdateView(en: boolean);
    var
    Viewer: TATBinHex;

  end;

var
  FormView: TFormView;

implementation

uses
  ATxSProc, ATxFProc, ATxCodepages;

{$R *.DFM}

procedure TFormView.chkModeTextClick(Sender: TObject);
begin
  if not FLoad then
    begin
    Viewer.Mode:= vbmodeText;
    UpdateView(true);
    end;
end;

procedure TFormView.chkModeBinaryClick(Sender: TObject);
begin
  if not FLoad then
    begin
    Viewer.Mode:= vbmodeBinary;
    UpdateView(true);
    end;
end;

procedure TFormView.chkModeHexClick(Sender: TObject);
begin
  if not FLoad then
    begin
    Viewer.Mode:= vbmodeHex;
    UpdateView(true);
    end;
end;

procedure TFormView.chkModeUnicodeClick(Sender: TObject);
begin
  if not FLoad then
    begin
    Viewer.Mode:= vbmodeUnicode;
    UpdateView(true);
    end;
end;

procedure TFormView.chkModeUHexClick(Sender: TObject);
begin
  if not FLoad then
    begin
    Viewer.Mode:= vbmodeUHex;
    UpdateView(true);
    end;
end;

procedure TFormView.UpdateView(en: boolean);
begin
  FLoad:= true;

  chkModeText.Checked:= Viewer.Mode = vbmodeText;
  chkModeBinary.Checked:= Viewer.Mode = vbmodeBinary;
  chkModeHex.Checked:= Viewer.Mode = vbmodeHex;
  chkModeUnicode.Checked:= Viewer.Mode = vbmodeUnicode;
  chkModeUHex.Checked:= Viewer.Mode = vbmodeUHex;

  chkModeText.Enabled:= en;
  chkModeBinary.Enabled:= en;
  chkModeHex.Enabled:= en;
  chkModeUnicode.Enabled:= en;
  chkModeUHex.Enabled:= en;

  chkWordWrap.Enabled:= en and (Viewer.Mode in [vbmodeText, vbmodeUnicode]);
  chkOEM.Enabled:= en and not (Viewer.Mode in [vbmodeUnicode, vbmodeUHex]);
  chkEnabled.Enabled:= en;
  chkNonPrint.Enabled:= en;

  FLoad:= false;
end;

procedure TFormView.ViewerDrawLine(ASender: TObject; ACanvas: TCanvas;
  const AStr, AStrAll: WideString; const ARect: TRect; const ATextPnt: TPoint;
  var ADone: Boolean);
begin
  // test
end;

procedure TFormView.FormShow(Sender: TObject);
var
  fn: WideString;
begin
  UpdateView(false);
end;

procedure TFormView.Open1Click(Sender: TObject);
var
OK: boolean;
begin
  OK := false;
  if OpenDialog1.Execute then
   OK:= Viewer.Open(OpenDialog1.FileName);
  Viewer.Visible:= OK;
  UpdateView(OK);

end;

procedure TFormView.FormCreate(Sender: TObject);
begin
  {$ifdef STREAMS}
  FStream:= nil;
  {$endif}
  FLoad:= false;
  Viewer := TATBinHex.Create(self);
  Viewer.Parent := self;
  Viewer.Align := alClient;
//  TreeEx.VirtualExplorerListview:= LVEx; //For compatability with VST2.0 installed in IDE
end;

procedure TFormView.FormDestroy(Sender: TObject);
begin
  {$ifdef STREAMS}
  if Assigned(FStream) then
    begin
    Viewer.OpenStream(nil);
    FStream.Free;
    FStream:= nil;
    end;
  {$endif}
end;

procedure TFormView.chkOEMClick(Sender: TObject);
begin
  if chkOEM.Checked then
    Viewer.TextEncoding:= vencOEM
  else
    Viewer.TextEncoding:= vencANSI;
end;

procedure TFormView.chkWordWrapClick(Sender: TObject);
begin
  Viewer.TextWrap:= chkWordWrap.Checked;
end;

procedure TFormView.chkEnabledClick(Sender: TObject);
begin
  Viewer.Enabled:= chkEnabled.Checked;
end;

procedure TFormView.chkNonPrintClick(Sender: TObject);
begin
  Viewer.TextNonPrintable:= chkNonPrint.Checked;
end;

end.
