
#include <vcl.h>
#pragma hdrstop

#include "ATFileNotification.h"



/* Helper thread class */


#include "d2c_system.h"


struct TATBinHexFileRec;
class TDirThread;




DWORD cATFileNotifyFlags [ 8 /*# TATFileNotifyOption */ ];
void cATFileNotifyFlagsInit( )
{
  cATFileNotifyFlags[0] = FILE_NOTIFY_CHANGE_FILE_NAME;
  cATFileNotifyFlags[1] = FILE_NOTIFY_CHANGE_DIR_NAME;
  cATFileNotifyFlags[2] = FILE_NOTIFY_CHANGE_ATTRIBUTES;
  cATFileNotifyFlags[3] = FILE_NOTIFY_CHANGE_SIZE;
  cATFileNotifyFlags[4] = FILE_NOTIFY_CHANGE_LAST_WRITE;
  cATFileNotifyFlags[5] = FILE_NOTIFY_CHANGE_LAST_ACCESS;
  cATFileNotifyFlags[6] = FILE_NOTIFY_CHANGE_CREATION;
  cATFileNotifyFlags[7] = FILE_NOTIFY_CHANGE_SECURITY;
}
AnsiString sMsgNotifError = "Error";
AnsiString sMsgNotifExceptionWait = "Exception while waiting for notification";
AnsiString sMsgNotifExceptionCreate = "Exception while creating thread";
AnsiString sMsgNotifExceptionTerminate = "Exception while terminating thread";




class TDirThread: public TThread {
  typedef TThread inherited;
  private:
  WideString prDirectory;
  WideString prFileName;
  void* prKillEvent;
  bool prSubtree;
  DWORD prNotifyFilter;
  TTimer* prTimer;
  protected:
  virtual void __fastcall Execute( );
  public:
  __fastcall TDirThread( const WideString ADirectory, const WideString AFileName, bool ASubtree, DWORD ANotifyFilter, TTimer* ATimer );
  __fastcall virtual ~TDirThread( );
  public: inline __fastcall TDirThread( bool CreateSuspended ) : inherited( CreateSuspended ) {}
};


/* Helper functions */




struct TATBinHexFileRec {
  bool FExist;
  DWORD FSizeLow, FSizeHigh;
  DWORD FAttr;
  TFileTime FTimeWr, FTimeCr, FTimeAcc;
};



void __fastcall FGetFileRec( const WideString Filename, TATBinHexFileRec& Rec )
{
  void* h = 0;
  TWin32FindDataA fdA;
  TWin32FindDataW fdW;
  FillChar( &Rec, sizeof( Rec ), 0 );
  if ( Win32Platform == VER_PLATFORM_WIN32_NT )
  {
    h = FindFirstFileW( Filename.c_bstr(), &fdW );
    Rec.FExist = h != INVALID_HANDLE_VALUE;
    if ( Rec.FExist )
    {
      Rec.FSizeLow = fdW.nFileSizeLow;
      Rec.FSizeHigh = fdW.nFileSizeHigh;
      Rec.FAttr = fdW.dwFileAttributes;
      Rec.FTimeWr = fdW.ftLastWriteTime;
      Rec.FTimeCr = fdW.ftCreationTime;
      Rec.FTimeAcc = fdW.ftLastAccessTime;
      /*# Windows */::FindClose( h );
    }
  }
  else
  {
    h = FindFirstFileA( ((AnsiString) Filename ).c_str(), &fdA );
    Rec.FExist = h != INVALID_HANDLE_VALUE;
    if ( Rec.FExist )
    {
      Rec.FSizeLow = fdA.nFileSizeLow;
      Rec.FSizeHigh = fdA.nFileSizeHigh;
      Rec.FAttr = fdA.dwFileAttributes;
      Rec.FTimeWr = fdA.ftLastWriteTime;
      Rec.FTimeCr = fdA.ftCreationTime;
      Rec.FTimeAcc = fdA.ftLastAccessTime;
      /*# Windows */::FindClose( h );
    }
  }
}


bool __fastcall FTimesDif( const TFileTime& Time1, const TFileTime& Time2 )
{
  bool result = false;
  result = ( Time1.dwLowDateTime != Time2.dwLowDateTime ) || ( Time1.dwHighDateTime != Time2.dwHighDateTime );
  return result;
}


bool __fastcall FFileChanged( const WideString Filename, DWORD Filter, TATBinHexFileRec& OldRec )
{
  bool result = false;
  TATBinHexFileRec NewRec;
  FGetFileRec( Filename, NewRec );
  result = ( OldRec.FExist != NewRec.FExist ) || ( ( ( ( Filter & FILE_NOTIFY_CHANGE_ATTRIBUTES ) ) != 0 ) && ( OldRec.FAttr != NewRec.FAttr ) ) || ( ( ( ( Filter & FILE_NOTIFY_CHANGE_SIZE ) ) != 0 ) && ( ( OldRec.FSizeLow != NewRec.FSizeLow ) || ( OldRec.FSizeHigh != NewRec.FSizeHigh ) ) ) || ( ( ( ( Filter & FILE_NOTIFY_CHANGE_LAST_WRITE ) ) != 0 ) && FTimesDif( OldRec.FTimeWr, NewRec.FTimeWr ) ) || ( ( ( ( Filter & FILE_NOTIFY_CHANGE_LAST_ACCESS ) ) != 0 ) && FTimesDif( OldRec.FTimeAcc, NewRec.FTimeAcc ) ) || ( ( ( ( Filter & FILE_NOTIFY_CHANGE_CREATION ) ) != 0 ) && FTimesDif( OldRec.FTimeCr, NewRec.FTimeCr ) );
  if ( result )
    Move( &NewRec, &OldRec, sizeof( TATBinHexFileRec ) );
  return result;
}


DWORD __fastcall FNotifyOptionsToFlags( TATFileNotifyOptions Options )
{
  DWORD result = 0;
  TATFileNotifyOption Opt;
  result = 0;
  for ( int stop = /*# High(TATFileNotifyOption) */ 7, Opt = /*# Low(TATFileNotifyOption) */ 0; Opt <= stop; Inc( Opt ) )
    if ( Options.Contains(Opt ) )
      result += cATFileNotifyFlags[Opt];
  return result;
}


void __fastcall MsgErr( const AnsiString S )
{
  MessageBoxA( 0, S.c_str(), sMsgNotifError.c_str(), MB_OK | MB_ICONERROR | MB_APPLMODAL );
}

/* Unicode versions of SysUtils' functions */


int __fastcall LastDelimiter( const WideString Delimiters, const WideString S )
{
  int result = 0;
  int i = 0;
  for ( int stop = 1, i = S.Length( ); i >= stop; i--)
    if ( Delimiters.Pos( WideString( S[i] ) ) > 0 )
    {
      result = i;
      return result;
    }
  result = 0;
  return result;
}


WideString __fastcall SExtractFileDir( const WideString Filename )
{
  WideString result;
  int i = 0;
  i = LastDelimiter( L"\\:", Filename );
  if ( ( i > 1 ) && ( Filename[i] == L'\\' ) )
    i--;
  result = Filename.SubString( 1, i );
  return result;
}

/*
function SExtractFilePath(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := LastDelimiter('\:', FileName);
  Result := Copy(FileName, 1, I);
end;
*/


/* TDirThread */


__fastcall TDirThread::TDirThread( const WideString ADirectory, const WideString AFileName, bool ASubtree, DWORD ANotifyFilter, TTimer* ATimer )
: inherited( false )
{
  prKillEvent = CreateEvent( NULL, false, false, NULL );
  prDirectory = ADirectory;
  prFileName = AFileName;
  prSubtree = ASubtree;
  prNotifyFilter = ANotifyFilter;
  prTimer = ATimer;
}


__fastcall TDirThread::~TDirThread( )
{
  SetEvent( prKillEvent );
  CloseHandle( prKillEvent );
  // todo check:  inherited::Destroy();
}


void __fastcall TDirThread::Execute( )
{
  void* ObjList[ 2/*# range 0..1*/ ];
  void* NotifyRes = 0;
  WideString ADir;
  bool ASubtree = false;
  DWORD AFilter = 0;
  TATBinHexFileRec AFileRec;
  bool IsFile = false;
  FillChar( &AFileRec, sizeof( TATBinHexFileRec ), 0 );
  IsFile = !prFileName.IsEmpty();
  if ( IsFile )
  {
    ADir = SExtractFileDir( prFileName );
    if ( ( !ADir.IsEmpty()) && ( ADir[ADir.Length( )] == L':' ) )
      ADir = ADir + L"\\"; //Handle the case of 'C:\Filename'
    ASubtree = false;
    AFilter = prNotifyFilter & ( ~ ( FILE_NOTIFY_CHANGE_DIR_NAME | FILE_NOTIFY_CHANGE_SECURITY ) );
    FGetFileRec( prFileName, AFileRec );
  }
  else
  {
    ADir = prDirectory;
    ASubtree = prSubtree;
    AFilter = prNotifyFilter;
  }

  //Create notification
  if ( Win32Platform == VER_PLATFORM_WIN32_NT )
    NotifyRes = FindFirstChangeNotificationW( ADir.c_bstr(), ASubtree, AFilter );
  else
    NotifyRes = FindFirstChangeNotificationA( ((AnsiString) ADir ).c_str(), ASubtree, AFilter );
  ObjList[0] = prKillEvent;
  ObjList[1] = NotifyRes;

  //Wait
  if ( NotifyRes != INVALID_HANDLE_VALUE )
    try
    {
      do
      {
        if //In some unknown circumstances signaling through prKillEvent may not work,
                       //so there is additional check for Terminated to stop thread
                       //during inherited TThread.Destroy.
        ( Terminated || ( WaitForMultipleObjects( 2, &ObjList[0], false, INFINITE ) == WAIT_OBJECT_0 ) )
        {
          break;
        }
        if ( ( ! IsFile ) || ( FFileChanged( prFileName, AFilter, AFileRec ) ) )
        {
          prTimer->Enabled = true;
        }
      }
      while ( ! ( ! FindNextChangeNotification( ObjList[1] ) ) );
      FindCloseChangeNotification( ObjList[1] );
    }
    catch(...)
    {
      MsgErr( sMsgNotifExceptionWait );
    }
}


/* TATFileNotification */


__fastcall TATFileNotification::TATFileNotification( TComponent* AOwner )
: inherited( AOwner )
{
  FStarted = false;
  FSubtree = false;
  FDirectory.Empty();
  FFileName.Empty();
  fOptions = ( TATFileNotifyOptions ()
                       << TATFileNotifyOption ( 0 ) << TATFileNotifyOption ( 1 ) << TATFileNotifyOption ( 4 ) );
  FTimer = new TTimer( this );
  /*# with FTimer do */
  {
    FTimer->Enabled = false;
    FTimer->Interval = 100;
    FTimer->OnTimer = FTimer->Timer;
  }
  InitializeCriticalSection( &FLock );
}


__fastcall TATFileNotification::~TATFileNotification( )
{
  if ( ! ( ComponentState.Contains(csDesigning ) ) )
  {
    Stop();
  }
  DeleteCriticalSection( &FLock );
  // todo check:  inherited::Destroy();
}


void __fastcall TATFileNotification::Start( )
{
  try
  {
    EnterCriticalSection( &FLock );
    try
    {
      if ( ! FStarted )
      {
        FDirThread = new TDirThread( FDirectory, FFileName, FSubtree, FNotifyOptionsToFlags( fOptions ), FTimer );
        FStarted = true;
      }
    }
    __finally
    {
      LeaveCriticalSection( &FLock );
    }
  }
  catch(...)
  {
    MsgErr( sMsgNotifExceptionCreate );
  }
}


void __fastcall TATFileNotification::Stop( )
{
  try
  {
    EnterCriticalSection( &FLock );
    try
    {
      if ( FStarted )
      {
        if (( FDirThread != NULL ) )
        {
          delete FDirThread;
          FDirThread = NULL;
        }
        FStarted = false;
      }
    }
    __finally
    {
      LeaveCriticalSection( &FLock );
    }
  }
  catch(...)
  {
    MsgErr( sMsgNotifExceptionTerminate );
  }
}


void __fastcall TATFileNotification::Timer( TObject* Sender )
{
  FTimer->Enabled = false;
  if (( FOnChanged != NULL ) )
    FOnChanged( this );
}


void __fastcall TATFileNotification::SetDirectory( const WideString ADirectory )
{
  if ( ADirectory != FDirectory )
  {
    FDirectory = ADirectory;
    FFileName.Empty();
  }
}


void __fastcall TATFileNotification::SetFileName( const WideString AFileName )
{
  if ( AFileName != FFileName )
  {
    FDirectory.Empty();
    FFileName = AFileName;
  }
}


void __fastcall TATFileNotification::SetEnabled( bool AValue )
{
  if ( AValue != FStarted )
  {
    if ( AValue )
      Start();
    else
      Stop();
  }
}

/* Registration */


void __fastcall Register( )
{
  TComponentClass classes[ 2 ] = { __classid( TATFileNotification) };
  RegisterComponents( "Samples" , classes,  1 );
}

class ATFileNotification_unit
{
public:
ATFileNotification_unit()
{
  cATFileNotifyFlagsInit();
}
};
ATFileNotification_unit _ATFileNotification_unit;
