
/**************************************************/
/*                                                */
/*  ATFileNotificationSimple Component            */
/*  Copyright (C) 2007 Alexey Torgashin           */
/*  http://atorg.net.ru                           */
/*                                                */
/**************************************************/
/*
ATFileNotificationSimple is a simple version of ATFileNotification,
the advantages are:
- It doesn't create any threads, it uses simple timer instead
- It doesn't use FindFirstChangeNotification API
- network files can be watched

Disadvantages are:
- It monitors file change only, not directory change
- It doesn't react to file change immediately (but with a timer delay)
- It polls disk repeatedly, may be slow on remote disks

Example of usage:
  procedure TFormMain.NotifyFile;
  begin
    with ATFileNotificationSimple1 do
    begin
      Timer.Enabled := False;
      Timer.Interval := StrToIntDef(edDelay.Text, Timer.Interval);
      FileName := edFileName.Text;
      Timer.Enabled := True;
    end;
  end;
*/

/*$BOOLEVAL OFF*/ //Short boolean evaluation.
#ifndef ATFileNotificationSimpleH
#define ATFileNotificationSimpleH


#include "d2c_system.h"

#include <windows.hpp>
#include <sysutils.hpp>
#include <classes.hpp>
#include <extctrls.hpp>



class TATFileNotificationSimple;
struct TATFileSimpleRec;




struct TATFileSimpleRec {
  bool FExist;
  DWORD FSizeLow, FSizeHigh;
  TFileTime FTimeWr;
};



class TATFileNotificationSimple: public TComponent {
  typedef TComponent inherited;
  private:
    /* Private declarations */
  WideString FFileName;
  TATFileSimpleRec FFileRec;
  TTimer* FTimer;
  bool FTimerBusy;
  TNotifyEvent FOnChanged;
  void __fastcall SetFileName( const WideString AValue );
  void __fastcall TimerTimer( TObject* Sender );
  void __fastcall DoChanged( );
  protected:
    /* Protected declarations */
  public:
    /* Public declarations */
  __fastcall TATFileNotificationSimple( TComponent* AOwner );
  __property WideString Filename = { read = FFileName, write = SetFileName };
  __property TTimer* Timer = { read = FTimer };
  __published:
    /* Published declarations */
  __property TNotifyEvent OnChanged = { read = FOnChanged, write = FOnChanged };
};


void __fastcall Register( );

#endif //  ATFileNotificationSimpleH