
#include <vcl.h>
#pragma hdrstop

#include "ATStreamSearch.h"



#include <sysutils.hpp>
#include "ATxSProc.h"



#include "d2c_system.h"


AnsiString MsgATStreamSearchRegExError = "Regular expression pattern error:\x0d\x0a\x0d\x0a" "%s at offset %d";
AnsiString MsgATStreamSearchReadError = "Read error at offset %d";



/* Constants */

const int cBlockSize = 64 * 1024;

/* Helper functions */


int __fastcall CharSize( TATEncoding AEncoding )
{
  int result = 0;
  if ( cATUnicodeEncodings.Contains(AEncoding ) )
    result = 2;
  else
    result = 1;
  return result;
}


int __fastcall BoolToSign( bool AValue )
{
  int result = 0;
  if ( AValue )
    result = 1;
  else
    result = - 1;
  return result;
}


void __fastcall NormalizePos( __int64& APos, int ACharSize )
{
  if ( ACharSize != 1 )
    APos = APos / ACharSize * ACharSize;
}


__int64 __fastcall LastPos( const __int64 AFileSize, int ACharSize )
{
  __int64 result = 0;
  result = AFileSize;
  NormalizePos( result, ACharSize );
  result -= ACharSize;
  I64LimitMin( result, 0 );
  return result;
}

/* TATStreamSearch */


__fastcall TATStreamSearch::TATStreamSearch( TComponent* AOwner )
: inherited( AOwner )
{
  FStream = NULL;
  FStreamOwner = false;
  FFileName.Empty();
  FStreamStart = - 1;
  FStreamSize = 0;
  FFoundStart = - 1;
  FFoundLength = 0;
  //FSearchForValidUTF16 := False;
  FOnProgress = NULL;
  FCharSize = 1;
  InitSavedOptions();
}


__fastcall TATStreamSearch::~TATStreamSearch( )
{
  FreeStream();
  // todo check:  inherited::Destroy();
}


void __fastcall TATStreamSearch::FreeStream( )
{
  if ( FStreamOwner )
    if (( FStream != NULL ) )
    {
      delete FStream;
      FStream = NULL;
    }
}


void __fastcall TATStreamSearch::InitSavedOptions( )
{
  FSavedText.Empty();
  FSavedEncoding = vEncANSI;
  FSavedOptions = TATStreamSearchOptions ();
}


void __fastcall TATStreamSearch::SetFileName( const WideString AFileName )
{
  FreeStream();
  if ( !AFileName.IsEmpty())
  {
    InitSavedOptions();
    FFileName = AFileName;
    FStreamOwner = true;
    FStream = new TFileStream( AFileName, fmOpenRead | fmShareDenyNone );
  }
}


void __fastcall TATStreamSearch::SetStream( TStream* AStream )
{
  FreeStream();
  InitSavedOptions();
  FFileName.Empty();
  FStreamOwner = false;
  FStream = AStream;
}


bool __fastcall TATStreamSearch::InitProgressFields( const __int64 AStartPos, TATEncoding AEncoding )
{
  bool result = false;
  FStreamStart = AStartPos;
  FStreamSize = FStream->Size;
  FCharSize = CharSize( AEncoding );
  result = FStreamSize >= FCharSize;
  return result;
}


void __fastcall TATStreamSearch::DoProgress( const __int64 ACurrentPos, const __int64 AMaximalPos, bool& AContinueSearching )
{
  AContinueSearching = true;
  if (( FOnProgress != NULL ) )
    FOnProgress( ACurrentPos, AMaximalPos, AContinueSearching );
}

//-----------------------------------------------------------------
// RegEx-related code
  

//-----------------------------------------------------------------
// Plain search code


__int64 __fastcall TATStreamSearch::TextFind( const WideString AText, const __int64 AStartPos, TATEncoding AEncoding, TATStreamSearchOptions AOptions )
{
  __int64 result = 0;
  char Buffer[ 65536/*# range 0..cBlockSize-1*/ ];
  __int64 BufPosMax = 0, BufPos = 0, ReadPos = 0;
  DWORD ReadSize = 0, BytesRead = 0;
  AnsiString SBufferA;
  WideString SBufferW;
  int StringPos = 0;
  bool AForward = false, AWholeWords = false, ACaseSens = false, AContinue = false;
  result = - 1;
  if ( AText.IsEmpty())
    return result;

  //1. Init objects and fields
  Assert(( FStream != NULL ), "Stream object not initialized" );
  if ( ! InitProgressFields( AStartPos, AEncoding ) )
    return result;

  //2. Init variables
  AForward = ! ( AOptions.Contains(asoBackward ) );
  AWholeWords = AOptions.Contains(asoWholeWords );
  ACaseSens = AOptions.Contains(asoCaseSens );
  BufPosMax = LastPos( FStreamSize, FCharSize );
  NormalizePos( BufPosMax, FCharSize );
  BufPos = AStartPos;
  NormalizePos( BufPos, FCharSize );
  if ( BufPos > BufPosMax )
  {
    if ( AForward )
      return result;
    else
      BufPos = BufPosMax;
  }
  if ( BufPos < 0 )
  {
    if ( AForward )
      BufPos = 0;
    else
      return result;
  }

  //3. Search
  DoProgress( BufPos, FStreamSize, AContinue );
  if ( ! AContinue )
    return result;
  do
  {
    ReadPos = BufPos;
    ReadSize = cBlockSize;
    if ( ! AForward )
    {
      ReadPos -= cBlockSize - FCharSize;
      I64LimitMin( ReadPos, 0 );
      ReadSize = BufPos - ReadPos + FCharSize;
      if ( ReadSize > cBlockSize )
        ReadSize = cBlockSize;
    }
    try
    {
      FillChar( Buffer, sizeof( Buffer ), 0 );
      FStream->Position = ReadPos;
      BytesRead = FStream->Read( Buffer, ReadSize );
    }
    catch(...)
    {
      throw Exception( Format( MsgATStreamSearchReadError, ARRAYOFCONST(( ReadPos )) ) );
      return result;
    }
    if ( FCharSize == 2 )
    {
      SBufferW = SetStringW( Buffer, BytesRead, AEncoding == vEncUnicodeBE );
      StringPos = SFindTextW( AText, SBufferW, AForward, AWholeWords, ACaseSens, BytesRead < cBlockSize );
    }
    else
    {
      SetString( SBufferA, Buffer, BytesRead );
      SBufferA = SCodepageToUnicode( SBufferA, AEncoding );
      StringPos = SFindText( AText, SBufferA, AForward, AWholeWords, ACaseSens, BytesRead < cBlockSize );
    }
    if ( StringPos > 0 )
    {
      result = ReadPos + ( StringPos - 1 ) * FCharSize;
      return result;
    }
    DoProgress( BufPos, FStreamSize, AContinue );
    if ( ! AContinue )
      return result;
    BufPos += ((__int64) ReadSize ) * BoolToSign( AForward );
    BufPos -= ((__int64) ( AText.Length( ) + 1 ) ) * FCharSize * BoolToSign( AForward );
    NormalizePos( BufPos, FCharSize );
    if ( ( BufPos < 0 ) || ( BufPos > BufPosMax ) )
      return result;
  }
  while ( ! ( BytesRead < cBlockSize ) );
  return result;
}


bool __fastcall TATStreamSearch::TextFindFirst( const WideString AText, const __int64 AStartPos, TATEncoding AEncoding, TATStreamSearchOptions AOptions )
{
  bool result = false;
  __int64 ARealStartPos = 0;
  ARealStartPos = AStartPos;
  if ( ( AOptions.Contains(asoBackward ) ) && ( AStartPos == 0 ) )
	ARealStartPos = 9223372036854775807; // High< __int64 >(); doesn't exist in the demo code
  FFoundStart = TextFind( AText, ARealStartPos, AEncoding, AOptions );
  result = FFoundStart >= 0;
  if ( result )
    FFoundLength = AText.Length( ) * FCharSize;
  else
    FFoundLength = 0;
  return result;
}


bool __fastcall TATStreamSearch::TextFindNext( bool AFindPrevious )
{
  bool result = false;
  int ACharSize = 0;
  __int64 AStartPos = 0, AForwardPos = 0, ABackwardPos = 0;
  TATStreamSearchOptions ANewOptions;
  ACharSize = CharSize( FSavedEncoding );
  AForwardPos = FFoundStart + ACharSize;
  ABackwardPos = FFoundStart + ( FSavedText.Length( ) - 2 ) * ACharSize;
  if ( ( FSavedOptions.Contains(asoBackward ) ) ^ ( AFindPrevious ) )
    AStartPos = ABackwardPos;
  else
    AStartPos = AForwardPos;
  ANewOptions = FSavedOptions;
  if ( AFindPrevious )
    if ( ANewOptions.Contains(asoBackward ) )
      ANewOptions >> asoBackward;
    else
      ANewOptions << asoBackward;
  FFoundStart = TextFind( FSavedText, AStartPos, FSavedEncoding, ANewOptions );
  result = FFoundStart >= 0;
  if ( result )
    FFoundLength = FSavedText.Length( ) * ACharSize;
  else
    FFoundLength = 0;
  return result;
}

//-----------------------------------------------------------------
// Combined search code


bool __fastcall TATStreamSearch::FindFirst( const WideString AText, const __int64 AStartPos, TATEncoding AEncoding, TATStreamSearchOptions AOptions )
{
  bool result = false;
  InitSavedOptions();
  FSavedText = AText;
  FSavedEncoding = AEncoding;
  FSavedOptions = AOptions;
  result = TextFindFirst( AText, AStartPos, AEncoding, AOptions );
  return result;
}


bool __fastcall TATStreamSearch::FindNext( bool AFindPrevious )
{
  bool result = false;
  Assert( !FSavedText.IsEmpty(), "Search text is empty" );
  result = TextFindNext( AFindPrevious );
  return result;
}


/* Registration */


void __fastcall Register( )
{
  TComponentClass classes[ 2 ] = { __classid( TATStreamSearch) };
  RegisterComponents( "Samples" , classes,  1 );
}
